package net.sergeych.crypto2

import net.sergeych.bintools.KVStorage
import net.sergeych.bintools.MemoryKVStorage
import net.sergeych.bintools.optStored
import net.sergeych.synctools.ProtectedOp
import net.sergeych.synctools.invoke
import kotlin.random.Random
import kotlin.random.nextUBytes

/**
 * Encrypted variant of [KVStorage]; the storage is encrypted with the given key
 * in a given [plainStore] [KVStorage]. It is threadsafe where
 * applicable. Also, it supports in-place key change [reEncrypt].
 *
 * Keys are stored encrypted and used hashed so it is not possible to
 * retrieve them without knowing the encryption key.
 *
 * @param plainStore where to store encrypted data
 * @param encryptionKey key to decrypt existing/encrypt new data. Can cause [DecryptionFailedException]
 *          if the key is wrong and the storage is already initialized with a new key and same [prefix]
 * @param prefix prefix for keys to distinguish from other data in [plainStore]
 * @param removeExisting if true, removes all existing data in [plainStore] if the [encryptionKey] can't
 *          decrypt existing encrypted data
 */
class EncryptedKVStorage(
    private val plainStore: KVStorage,
    private var encryptionKey: SymmetricKey,
    private val prefix: String = "EKVS_",
    removeExisting: Boolean
) : KVStorage {
    private val op = ProtectedOp()

    private val prefix2 = prefix + ":"

    val seed: UByteArray

    init {
        var encryptedSeed by plainStore.optStored<UByteArray>("$prefix#seed")
        seed = try {
            encryptedSeed?.let { encryptionKey.decrypt(it) }
                ?: Random.nextUBytes(32).also {
                    encryptedSeed = encryptionKey.encrypt(it)
                }
        } catch (x: DecryptionFailedException) {
            if (removeExisting) {
                plainStore.keys.filter { it.startsWith(prefix) }.forEach {
                    plainStore.delete(it)
                }
                Random.nextUBytes(32).also {
                    encryptedSeed = encryptionKey.encrypt(it)
                }
            } else throw x
        }
    }

    private fun mkkey(key: String): String =
        blake2b(key.encodeToByteArray().asUByteArray() + seed).encodeToBase64Url()

    override val keys: Set<String>
        get() = op.invoke {
            plainStore.keys.mapNotNull {
                if (it.startsWith(prefix2))
                    plainStore[it]?.let { encrypted ->
                        encryptionKey.decrypt(encrypted.asUByteArray()).asByteArray().decodeToString()
                    }
                else null
            }.toSet()
        }

    override fun get(key: String): ByteArray? = op {
        val k0 = mkkey(key)
        val k = prefix + k0
        plainStore[k]?.let { encryptionKey.decrypt(it.asUByteArray()).asByteArray() }
            ?.also {
                val k2 = prefix2 + k0
                if (k2 !in plainStore)
                    plainStore[k2] = encryptionKey.encrypt(key).asByteArray()
            }
    }

    override fun set(key: String, value: ByteArray?) {
        op {
            val k1 = mkkey(key)
            plainStore[prefix + k1] = value?.let {
                encryptionKey.encrypt(it.asUByteArray()).asByteArray()
            }
            plainStore[prefix2 + k1] = encryptionKey.encrypt(key).asByteArray()
        }
    }

    /**
     * Re-encrypts the entire storage in-place with the given key; it is threadsafe where
     * applicable.
     *
     * This method re-encrypts every data item so it is cryptographically secure.
     */
    fun reEncrypt(newKey: SymmetricKey) {
        op {
            val copy = MemoryKVStorage().also { it.addAll(this) }
            encryptionKey = newKey
            addAll(copy)
        }
    }
}